/*
 *
 * Normal mappings of chips in physical memory.
 *
 * Copyright (C) 2005,  Freescale Semiconductor (Matt.Waddel@freescale.com)
 * Copyright (C) 2005,  Intec Automation Inc. (mike@steroidmicros.com)
 * Copyright (C) 2001-2002, David McCullough <davidm@snapgear.com>
 *
 * Based on snapgear-uc.c
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/kdev_t.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <linux/root_dev.h>
#include <linux/dBug-config.h>
#include <linux/mm.h>

/*
 * Define the end of the kernel partition.
 *
 * The remainder of the bank will be allocated to the "UserFile"
 * partition and that will contain a JFFS.  Since JFFS requires a
 * minimum of five flash-blocks, slightly more than that is reserved.
 *
 * For instance: 0x800000 - (6*128k or 0xc0000) = 0x740000.
 */
#define KERNEL_END_ADDR 0x740000
//#define KERNEL_END_ADDR 0x400000

#define FLASH_BASE_ADDR 0x00000000

#define DEFAULT_FLASH_SIZE 0x00800000	// Matchport AR's default flash size is 8MB
#define BANKWIDTH 2

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *mymtd;

static struct mtd_info *ram_mtdinfo;
static struct map_info m520x_ram_map = {
	.name		= "RAM",
};
static struct mtd_partition m520x_romfs[] = {
	{
		.name	= "Romfs"
	}
};

/* to overwrite hardcoded partition layout set kcl rootfstype=romfs mtdparts=mpt:128k@128k(Lbootloader),3840k(LKernel),4096k(UserSpace) in dBug */

#ifdef CONFIG_MTD_CMDLINE_PARTS
static struct mtd_partition *mtd_parts;
static int                   mtd_parts_nb;
const char *part_probes[] = {"cmdlinepart", NULL};
#endif

static struct mtd_info *flash_mtdinfo;
struct map_info m520x_flash_map = {
	.name		= "mpt",
	.size		= DEFAULT_FLASH_SIZE,
	.bankwidth	= BANKWIDTH
};
static struct mtd_partition m520x_partitions_including_kernel[] = {
        {
                .name	= "LTRXbloader",
                .size	= 0x10000,
                .offset	= 0x0,
                .mask_flags	= MTD_WRITEABLE	/* force read-only */
        },
        {
                .name	= "LTRXconfig",
                .size	= 0x10000,
                .offset	= MTDPART_OFS_NXTBLK,
                .mask_flags	= MTD_WRITEABLE	/* force read-only */
        },
        {
                .name	= "dBug",
                .size	= 0x10000,
                .offset	= MTDPART_OFS_NXTBLK,
                .mask_flags	= MTD_WRITEABLE	/* force read-only */
        },
        {
                .name	= "dBugConfig",
                .size	= 0x10000,
                .offset	= MTDPART_OFS_NXTBLK,
        },
        {
                .name	= "Kernel",
                .size	= (DEFAULT_FLASH_SIZE / 2) - 0x40000,
                .offset	= MTDPART_OFS_NXTBLK,
                .mask_flags	= MTD_WRITEABLE	/* force read-only */
        },
};

static struct mtd_partition m520x_romfs_in_flash_partition[] = {
        {
                .name	= "Romfs",
                .size	= 0,  /* needs to be adjusted dynamically */
                .offset	= 0,  /* needs to be adjusted dynamically */
                .mask_flags	= MTD_WRITEABLE	/* force read-only */
        },
};

static struct mtd_partition m520x_partitions_after_romfs[] = {
        {
                .name	= "UserSpace",
                .size	= MTDPART_SIZ_FULL,   // gets overwritten below
                .offset	= MTDPART_OFS_NXTBLK
        },
        {
                .name	= "WritableFlash",
                .size	= MTDPART_SIZ_FULL,
                .offset	= 0x40000
        }
#ifdef CONFIG_MTD_RECOVER_PARAMS
        ,
        {
                .name	= "UserExtra",
                .size	= 0x30000,
                .offset	= 0x400000  // gets overwritten below
        },
        {
                .name	= "dBugConfBackup",
                .size	= 0x10000,
                .offset	= 0x430000  // gets overwritten below
        }
#else /* CONFIG_MTD_RECOVER_PARAMS */
#ifdef CONFIG_MTD_BOOT_BANK
        ,
        {
                .name	= "UserExtra",
                .size	= 0x40000,
                .offset	= 0x400000  // gets overwritten below
        }
#endif /* CONFIG_MTD_BOOT_BANK */
#endif /* CONFIG_MTD_RECOVER_PARAMS */
};



/****************************************************************************
 *
 * Find the MTD device with the given name
 *
 ****************************************************************************/

static struct mtd_info
*get_mtd_named(char *name)
{
	int i;
	struct mtd_info *mtd;

	for (i = 0; i < MAX_MTD_DEVICES; i++) {
		mtd = get_mtd_device(NULL, i);

		if (!IS_ERR(mtd)) {
			if (strcmp(mtd->name, name) == 0)
				return(mtd);
			put_mtd_device(mtd);
		}
	}
	return(NULL);
}


static int
m520x_point(struct mtd_info *mtd, loff_t from, size_t len,
            size_t *retlen, void **mtdbuf, resource_size_t *phys)
{
	struct map_info *map = (struct map_info *) mtd->priv;
	*mtdbuf = (u_char *) (map->map_priv_1 + (int)from);
	*retlen = len;
	return(0);
}

#ifdef CONFIG_MTD_RECOVER_PARAMS
u_int32_t __init
m520x_get_dBugConfig(u_int32_t base)
{
	unsigned int p;
	unsigned int latest = 0;

	if (memcmp((void *)(base + PARAMBLK_SIGNATURE), PARAM_UNUSED, 8) == 0) {
		//printk("m520x_get_dBugConfig: Old config found\n");
		return base;
	}

	for (p = base; p < (base + (mymtd->size == 0x01000000 ? 0x20000 : 0x10000)); p += PARAMBLK_SIZE) {
		if (memcmp((void *)(p + PARAMBLK_SIGNATURE), PARAM_SIGNATURE, 8) == 0) {
			/* this block is valid */
			latest = p;
		}
	}

	//printk("m520x_get_dBugConfig: 0x%x / 0x%x (%d)\n", latest, base, (latest - base) / PARAMBLK_SIZE);
	return latest;
}
#endif /* CONFIG_MTD_RECOVER_PARAMS */

static int __init
m520x_probe(int ram, unsigned long addr, int size, int bankwidth)
{
	struct map_info *map_ptr;


	if (ram)
		map_ptr = &m520x_ram_map;
	else
		map_ptr = &m520x_flash_map;

	map_ptr->bankwidth = bankwidth;
	map_ptr->map_priv_2 = addr;
	map_ptr->phys = addr;
	map_ptr->size = size;

	printk(KERN_NOTICE "m520xevb %s probe(0x%lx,%x,%x): %lx at %lx\n",
			   ram ? "ram":"flash", addr, size, bankwidth,
			   map_ptr->size, map_ptr->map_priv_2);

	map_ptr->virt = ioremap_nocache(map_ptr->map_priv_2, map_ptr->size);

	simple_map_init(map_ptr);
	if (ram)
		mymtd = do_map_probe("map_ram", map_ptr);
	else
		mymtd = do_map_probe("cfi_probe", map_ptr);

	if (!mymtd) {
		iounmap((void *)map_ptr->map_priv_1);
		return -ENXIO;
	}

	if (!ram) {
		unsigned long real_flash_size = mymtd->size;
		m520x_flash_map.size = real_flash_size;
		switch(real_flash_size) {
			case 0x00400000:
				m520x_partitions_including_kernel[4].size = 0x200000 - 0x40000; // fix Kernel part
				break;

			case 0x01000000:
				m520x_partitions_including_kernel[0].size = 0x20000;  // fix LTRXbloader part
				m520x_partitions_including_kernel[1].size = 0x20000;  // fix LTRXconfig part
				m520x_partitions_including_kernel[2].size = 0x20000;  // fix dBUG part
				m520x_partitions_including_kernel[3].size = 0x20000;  // fix dBugConfig part
				m520x_partitions_including_kernel[4].size = KERNEL_END_ADDR - 0x80000; // fix Kernel part - size of bootloader partitions
				m520x_partitions_after_romfs[1].offset = 0x80000; // fix Writeable part
#ifdef CONFIG_MTD_RECOVER_PARAMS
				m520x_partitions_after_romfs[2].offset = (mymtd->size / 2);   // fix UserExtra part
				m520x_partitions_after_romfs[2].size = 0x60000;   // fix UserExtra part
				m520x_partitions_after_romfs[3].offset = (mymtd->size / 2) + 0x60000;   // fix dBugConfigBackup part
				m520x_partitions_after_romfs[3].size = 0x20000;   // fix dBugConfigBackup part
#else /* CONFIG_MTD_RECOVER_PARAMS */
#ifdef CONFIG_MTD_BOOT_BANK
				m520x_partitions_after_romfs[2].offset = (mymtd->size / 2);   // fix UserExtra part
				m520x_partitions_after_romfs[2].size = 0x80000;   // fix UserExtra part
#endif /* CONFIG_MTD_BOOT_BANK */
#endif /* CONFIG_MTD_RECOVER_PARAMS */
				break;
		}

	}

	mymtd->owner  = THIS_MODULE;
	mymtd->point  = m520x_point;
	mymtd->priv   = map_ptr;

	if (ram) {
		ram_mtdinfo = mymtd;
		add_mtd_partitions(mymtd, m520x_romfs, NB_OF(m520x_romfs));

	} else {
#ifdef CONFIG_MTD_CMDLINE_PARTS
		mtd_parts_nb = parse_mtd_partitions(mymtd, part_probes,
					        &mtd_parts, 0);
		printk(KERN_NOTICE "Found %d partitions on the command line.\n", mtd_parts_nb);

		if (mtd_parts_nb > 0)
		{
		    add_mtd_partitions (mymtd, mtd_parts, mtd_parts_nb);

		    return 0;
		}
#endif

#ifdef CONFIG_MTD_BOOT_BANK
		if (mymtd) {
			u_int32_t dbug_config_offset = 0;
			u_int32_t kernel_offset = 0;
			dBug_config_t *dbug_conf;
			u_int8_t bootbank;
			dbug_config_offset = m520x_partitions_including_kernel[0].size +
								 m520x_partitions_including_kernel[1].size +
								 m520x_partitions_including_kernel[2].size;
			kernel_offset = dbug_config_offset +
								 m520x_partitions_including_kernel[3].size;
#ifdef CONFIG_MTD_RECOVER_PARAMS
			dbug_conf = (dBug_config_t *)m520x_get_dBugConfig(dbug_config_offset);
#else
			dbug_conf = (dBug_config_t *)(dbug_config_offset);
#endif /* CONFIG_MTD_RECOVER_PARAMS */
			bootbank = dbug_conf->current_bank;

			/* Fix size for MatchPort AR */
			if ((mymtd->size == 0x00800000) &&
			    (DBUG_BOOTBANK_GET(bootbank) == 1 ||
			     DBUG_BOOTBANK_GET(bootbank) == 2)) {
				m520x_partitions_including_kernel[4].size = (DEFAULT_FLASH_SIZE / 4) - 0x40000; // fix Kernel part 2MB - size of bootloader partitions
			}

			/* Update kernel offset to 2nd bank */
			if (DBUG_BOOTBANK_GET(bootbank) == 2) {
				m520x_partitions_including_kernel[4].offset =
					kernel_offset + (mymtd->size / 2);
			}
		}
#endif /* CONFIG_MTD_BOOT_BANK */

		flash_mtdinfo = mymtd;
		add_mtd_partitions(mymtd, m520x_partitions_including_kernel, NB_OF(m520x_partitions_including_kernel));
	}
	return(0);
}


extern int _stext, _etext, _sdata, _edata, _sbss, _ebss, _end;
extern int _ramstart, _ramend;

/*
 * Initialize the mtd devices
 */

#define CSMR0_FOR_16MB     (0x00FF0001)
#define CSMR0_FOR_8MB      (0x007F0001)
#define CSMR0_FOR_4MB      (0x003F0001)

int __init init_m520x(void)
{
	int rc = -1;
	struct mtd_info *mtd;
	bool romfs_registered = false;
	unsigned int *CSMR0 = (unsigned int *) 0xfc008004;
	unsigned int flash_size;
	struct mtd_partition *kernel_partition = &m520x_partitions_including_kernel[4];

	/* guess actual flash size from chip select mask register - should have been set by dBUG correctly */
	switch(*CSMR0) {
		case CSMR0_FOR_4MB:
			flash_size = 4 * 1024 * 1024;
			break;
		case CSMR0_FOR_16MB:
			flash_size = 16 * 1024 * 1024;
			break;
		case CSMR0_FOR_8MB:
		default:
			flash_size = 8 * 1024 * 1024;
	}
	printk(KERN_NOTICE "Expected flash size: %u\n", flash_size);


	rc = m520x_probe( 0, FLASH_BASE_ADDR, flash_size, BANKWIDTH);
	/* Map in the filesystem from RAM last so that, if the filesystem
	 * is not in RAM for some reason we do not change the minor/major
	 * for the flash devices
	 */

	if (flash_mtdinfo) {
		u_int32_t offset = 0;
		u_int32_t dbug_config_offset = 0;
		u_int32_t kernel_offset = 0;
		bool dbug_config_found = false;
		bool kernel_found = false;
#ifdef CONFIG_MTD_RECOVER_PARAMS
		int partitions_after_romfs_size = NB_OF(m520x_partitions_after_romfs) - 2;
#else /* CONFIG_MTD_RECOVER_PARAMS */
#ifdef CONFIG_MTD_BOOT_BANK
		int partitions_after_romfs_size = NB_OF(m520x_partitions_after_romfs) - 1;
#endif	/* CONFIG_MTD_BOOT_BANK */
#endif /* CONFIG_MTD_RECOVER_PARAMS */

		struct mtd_info *mtd;
		int i;

		for (i = 0; i < MAX_MTD_DEVICES; i++) {
			mtd = get_mtd_device(NULL, i);

			if (IS_ERR(mtd)) {
//				printk(KERN_NOTICE "Next FS should be at 0x%x\n", offset);
				break;
			}

			if (strcmp(mtd->name, "Kernel") == 0) {
				uint64_t jffs_offset;
				KernelInFlashInfo_t *kernel_info;
				kernel_offset = offset;
#ifdef CONFIG_MTD_BOOT_BANK
				if (kernel_partition->offset != MTDPART_OFS_NXTBLK)
					kernel_offset = kernel_partition->offset;
#endif	/* CONFIG_MTD_BOOT_BANK */
				kernel_found = true;
				kernel_info = (KernelInFlashInfo_t *) kernel_offset;
				/* space for Kernel? */
				if (kernel_partition->size < kernel_info->image_size) {
					panic("0x%llx (%llu) byte Kernel partition too small for 0x%x (%u) byte kernel\n",
					      kernel_partition->size, kernel_partition->size,
					      kernel_info->image_size, kernel_info->image_size);
				}

				jffs_offset = kernel_offset + kernel_partition->size;
				m520x_partitions_after_romfs[0].offset = jffs_offset;
				printk(KERN_NOTICE "JFFS2 expected at 0x%llx\n", jffs_offset);

				break;
			} else if (strcmp(mtd->name, "dBugConfig") == 0) {
				dbug_config_found = true;
				dbug_config_offset = offset;
			}

//			printk(KERN_NOTICE "MTD %d: %s, 0x%x : 0x%x\n", i, mtd->name, mtd->size, mtd->erasesize);

			offset += mtd->size;


		}

		if (dbug_config_found) {
#ifdef CONFIG_MTD_RECOVER_PARAMS
			dBug_config_t *dbug_conf = (dBug_config_t *)m520x_get_dBugConfig(dbug_config_offset);
#else
			dBug_config_t *dbug_conf = (dBug_config_t *)dbug_config_offset;
#endif /* CONFIG_MTD_RECOVER_PARAMS */
#ifdef CONFIG_MTD_BOOT_BANK
			u_int8_t bootbank;
			int bank_size = mymtd->size / 2;
#endif	/* CONFIG_MTD_BOOT_BANK */

//			printk(KERN_NOTICE "dbug_config_offset = 0x%p.\n", dbug_conf);
			if (dbug_conf->flags & DBUG_ROMFS_FROM_FLASH_ENABLE) {
//				printk(KERN_NOTICE "Trying to mount ROMFS from flash.\n");
				if (kernel_found) {
					KernelInFlashInfo_t *kernel_info = (KernelInFlashInfo_t *) kernel_offset;
					u_int32_t romfs_addr = kernel_offset + kernel_info->image_size + 128;

//					printk(KERN_NOTICE "Looking at address 0x%x : 0x%x : %p : 0x%x for ROMFS.\n", romfs_addr, kernel_offset,  kernel_info, kernel_info->image_size);
					if (strncmp((char *) romfs_addr, "-rom1fs-", 8) == 0) {
						uint64_t kernel_plus_romfs_size = romfs_addr + kernel_info->romfs_size - kernel_offset;
						printk(KERN_NOTICE "Found ROMFS at address 0x%x\n", romfs_addr);
						if (kernel_partition->size < kernel_plus_romfs_size) {
							panic("0x%llx (%llu) byte kernel partition too small for 0x%llx (%llu) byte kernel+romfs\n",
							      kernel_partition->size, kernel_partition->size,
							      kernel_plus_romfs_size, kernel_plus_romfs_size);
						}
						m520x_romfs_in_flash_partition[0].offset = romfs_addr;
						m520x_romfs_in_flash_partition[0].size = kernel_info->romfs_size;
				                add_mtd_partitions(flash_mtdinfo, m520x_romfs_in_flash_partition, NB_OF(m520x_romfs_in_flash_partition));
				                romfs_registered = true;

					} else {
						printk(KERN_NOTICE "No ROMFS at address 0x%x.\n", romfs_addr);
					}
				}

			}
//			else
//				printk(KERN_NOTICE "Ignoring potential ROMFS in flash.\n");

#ifdef CONFIG_MTD_BOOT_BANK
			bootbank = dbug_conf->current_bank;
			if (DBUG_BOOTBANK_GET(bootbank) == 1 || DBUG_BOOTBANK_GET(bootbank) == 2 ||
			    mymtd->size == 0x01000000) {

				/* Add 'UserExtra' */
#ifdef CONFIG_MTD_RECOVER_PARAMS
				partitions_after_romfs_size += 2;
#else /* CONFIG_MTD_RECOVER_PARAMS */
				partitions_after_romfs_size++;
#endif /* CONFIG_MTD_RECOVER_PARAMS */
				/* UserSpace */
				if (m520x_partitions_after_romfs[0].offset < bank_size) {
					m520x_partitions_after_romfs[0].size = bank_size - m520x_partitions_after_romfs[0].offset;
				} else if (m520x_partitions_after_romfs[0].offset > bank_size) {
					m520x_partitions_after_romfs[0].size = mymtd->size - m520x_partitions_after_romfs[0].offset;
				} else {
					printk(KERN_INFO "Bank %s has zero size\n", m520x_partitions_after_romfs[0].name);
					m520x_partitions_after_romfs[0].size = 0;
				}
				/* WritableFlash */
				m520x_partitions_after_romfs[1].size = bank_size - m520x_partitions_after_romfs[1].offset;
				if (DBUG_BOOTBANK_GET(bootbank) == 1) {
					m520x_partitions_after_romfs[1].offset += bank_size;
				}
				/* UserExtra */
				if (DBUG_BOOTBANK_GET(bootbank) != 1 && DBUG_BOOTBANK_GET(bootbank) != 2) {
#ifdef CONFIG_MTD_RECOVER_PARAMS
					m520x_partitions_after_romfs[2].offset += 0x80000;
					m520x_partitions_after_romfs[2].size = bank_size - 0x80000;
#else /* CONFIG_MTD_RECOVER_PARAMS */
					m520x_partitions_after_romfs[2].size = bank_size;
#endif /* CONFIG_MTD_RECOVER_PARAMS */
				}
			}
#endif	/* CONFIG_MTD_BOOT_BANK */
		}

#ifdef CONFIG_MTD_BOOT_BANK
		add_mtd_partitions(flash_mtdinfo, m520x_partitions_after_romfs, partitions_after_romfs_size);
#else
		add_mtd_partitions(flash_mtdinfo, m520x_partitions_after_romfs, NB_OF(m520x_partitions_after_romfs));
#endif	/* CONFIG_MTD_BOOT_BANK */
	}


	if (!romfs_registered) {

		unsigned long start_area = (unsigned long) &_ebss;

		if (0 != m520x_probe(1, start_area,
				     PAGE_ALIGN(*(unsigned long *)(start_area + 8)), 4)) {
			printk(KERN_NOTICE "Failed to probe RAM filesystem\n");
		}

	}

	mtd = get_mtd_named("Romfs");
	if (mtd) {
		ROOT_DEV = MKDEV(MTD_BLOCK_MAJOR, mtd->index);
		put_mtd_device(mtd);
	} // else
//		printk(KERN_NOTICE "%s: Failed to make root filesystem\n", __FUNCTION__);

	return(rc);
}

static void __exit cleanup_m520x(void)
{
#ifdef CONFIG_MTD_CMDLINE_PARTS
	if (mtd_parts_nb) {
		del_mtd_partitions(mymtd);
		kfree(mtd_parts);
	} else {
#endif
		if (flash_mtdinfo) {
			del_mtd_partitions(flash_mtdinfo);
			map_destroy(flash_mtdinfo);
			flash_mtdinfo = NULL;
		}
#ifdef CONFIG_MTD_CMDLINE_PARTS
	}
#endif
	if (ram_mtdinfo) {
		del_mtd_partitions(ram_mtdinfo);
		map_destroy(ram_mtdinfo);
		ram_mtdinfo = NULL;
	}
	if (m520x_ram_map.map_priv_1) {
		iounmap((void *)m520x_ram_map.map_priv_1);
		m520x_ram_map.map_priv_1 = 0;
	}
	if (m520x_flash_map.map_priv_1) {
		iounmap((void *)m520x_flash_map.map_priv_1);
		m520x_flash_map.map_priv_1 = 0;
	}

}

module_init(init_m520x);
module_exit(cleanup_m520x);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("<Matt.Waddel@freescale.com>");
MODULE_DESCRIPTION("MTD map for M520xEVB");
